var labelType, useGradients, nativeTextSupport, animate;

/**
	Set up Error Text and Navigator
*/

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
    this.elem.style.left = (500 - this.elem.offsetWidth / 2) + 'px';
  }
};

		
function GraphFlow(){
	
	/////////////////////////////////////////////
	//				VARIABLES				   //
	/////////////////////////////////////////////
	
	var visContainer = 'infovis';
	
	/** Color Palettes */
	var nodeFillColor = [];
		nodeFillColor['nodeFamily']		= "#DEDEDE";
	
	var nodeStrokeColor = [];
		nodeStrokeColor['inputPort']   	  = "#AA0000";
		nodeStrokeColor['outputPort']	  = "#AA0000";
		nodeStrokeColor['repositoryPort'] = "#AA0000";
		nodeStrokeColor['nodeFamily'] 	  = "#4D4D4D";
		nodeStrokeColor['crossBox'] 	  = "#4D4D4D";
		
	var nodeColorFocus	  = "#0098BE",
		edgeColor 		  = "#114270",
		edgeColorFocus 	  = "#0098BE";
		
	/** Icon Images */
	var nodeIcons = [];
		
	/** Overall size modifier: the width of node labels in pixels */
	var vertexLabelSize 	= 12;

	/** FlowGraph Object */
	var fd;

	/** Nodes and edges where the mouse moves over */
	var hover = null;

	/** This array stores all Nodes and is initialized with a dummy,
		used for drawing edges between a node and the mousepointer.
		*/
	var json;
		
	/** Used for managing space for node creation/deletion
		since the array is divided into to parts, adding and removing nodes
		would cause many costly splices. We a similar approach to ArrayLists. 
	*/
	var jsonCapacity = {'occupied' : 0,
						'max'  : 10};

	/** selected node:	{id : <the id of the selected node, 
						 from : <the data.$type of the selected node>}
						 
		When clicking a node or edge, this variable will contain node
		information, that is required for creating edges.
		*/
	var selectedNode = null;
	
	/** This points to the DOM-node (not the json-array node!) that is
		an invisible dummy for moving an edge with the mouse
		*/
	var mouseNode = null;

	/** true if the mouse is over a port-node/edge */
	var mouseOverNode = false;
	var mouseOverEdge = false;
	
	/** A hashtable of mouse and node events to arrays of functions that
		should happen on these events
		*/
	var listener = [];
	
	/** Can be changed to allow or prohibit node movement or creation */
	var readOnly = {'deleteCreate' 	: true,
					'move' 			: true};
	
	/** Options concerning the grid. */
		
	var grid = { 'data': {'$dim' : vertexLabelSize*4,
						  '$type'	 : 'grid',
						  '$color'  : '#AAAAFF',
						  '$width' : document.getElementById(visContainer).clientWidth,
						  '$height': document.getElementById(visContainer).clientHeight
						},
				 'id':	'#DUMMY_GRID_NODE',
				 'visible': false,
				 'snap'   : false
				};
	
	/** Stores information about the canvas, such as its dimensions and its position
		*/
	var navi = { 'width' : document.getElementById(visContainer).clientWidth,
				 'height': document.getElementById(visContainer).clientHeight,
				 'centerX' : 0,
				 'centerY' : 0,
				 'dragX' : null,
				 'dragY' : null,
				 'draggedNodes' : null};
	
	/////////////////////////////////////////////
	//				FUNCTIONS				   //
	/////////////////////////////////////////////

	
	/**
		Changes the color of the grid. Requires a refresh().
	*/
	this.setGridColor = function(newColor){
		grid.data.$color = newColor;
	}
	
	/**
		Changes the size of the grid. Requires a refresh().
	*/
	this.setGridSize = function(newSize){
		grid.data.$dim = newSize;
	}
	
	/**
		Changes the visibility of the grid. Requires a refresh().
	*/
	this.setGridVisible = function(visibility){
		grid.visible = visibility;
	}
	
	/**
		Changes whether Nodes should snap to grid. Requires a refresh().
	*/
	this.setGridSnap = function(snap){
		grid.snap = snap;
	}
	
	
	/**
		Scales and centers the graph to fit the space provided by the container
		in which it is drawn.
	*/
	this.scaleToFit = function(){
	
	// do nothing if the graph is empty
	if(jsonCapacity.occupied == 0){
		return;
	}
	
	// init loop variables
	var data   = json[0].data,
		width  = data.$width/2,
		height = data.$height/2;
	
	var left  	= data.$xPos - width,
		right 	= data.$xPos + width, 
		top		= data.$yPos - height, 
		bottom	= data.$yPos + height;
	
	// init the outer bounds of the displayed graph
	var leftMost = left,
		rightMost = right,
		topMost = top,
		bottomMost = bottom;
	
	// calculate the outer bounds of the displayed graph
	for(var n = 1, l= jsonCapacity.occupied; n < l; n++){
		data   = json[n].data,
		width  = data.$width/2,
		height = data.$height/2;
	
		left  	= data.$xPos - width,
		right 	= data.$xPos + width, 
		top		= data.$yPos - height, 
		bottom	= data.$yPos + height;
		
		// update bounds
		if(left < leftMost){
			leftMost = left;
		}
		if(right > rightMost){
			rightMost = right;
		}
		if(top < topMost){
			topMost = top;
		}
		if(bottom > bottomMost){
			bottomMost = bottom;
		}
	}
	
	// scaling values
	var oldScale = fd.canvas.scaleOffsetX,
		scaleX = navi.width / (rightMost - leftMost + 4*vertexLabelSize),
		scaleY = navi.height / (bottomMost - topMost +4*vertexLabelSize)
		scale = Math.min(scaleX, scaleY) / oldScale;
		
	// translation values
	var midX = (leftMost + rightMost) / 2,
		midY = (topMost + bottomMost) / 2 -vertexLabelSize;
	
	// translate to origin, scale, and translate to the center of the graph
	fd.canvas.translate( -fd.canvas.translateOffsetX/oldScale, -fd.canvas.translateOffsetY/oldScale);
	fd.canvas.scale(scale, scale);
	fd.canvas.translate( -midX, -midY);
	}
	
	
	/**
	Assigns an iconPath to a type of node. Needs a refresh() to take effect.
	@param nodeType - the type of the node. Typically 'Repository',
					  'Filter', or 'Reader'
	@param path - the path to the image. If null, an existing
				  icon will be deleted
	*/
	this.setNodeIcon = function(nodeType, path){
		if(path == null){
			delete nodeIcons[nodeType];
			//TODO: adjust shrink width
		}
		else{
			// if nodes were already created, we need to
			// adjust their width and iconPath
			if(nodeIcons[nodeType] == undefined){
				var icon = Image();
					addSize = 6* vertexLabelSize,
					addPortSize = addSize / 2,
					max = jsonCapacity.max;
			
				icon.src = path;
				
				var data, portData, portCount, addPortSizeRel;
				for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
					// adjust node width
					data = json[n].data;
					if(data.$nodeType == nodeType){
						data.$icon = icon;
						data.$width += addSize;
						portCount = data.$portCount;
						
						// adjust port positions
						for(var p = data.$portIndex+max, s = p+portCount; p < s; p++){
							portData = json[p].data;
							if(portData.$type == "inputPort"){
								addPortSizeRel = -addPortSize;
							}else{
								addPortSizeRel = addPortSize;
							}
							if(portData.$relativeX != undefined){
								portData.$relativeX += addPortSizeRel;
							}
							portData.$xPos += addPortSizeRel;
						}
					}
				}
			}
			nodeIcons[nodeType] = icon;
		}
	}
	
	/**
	Sets the mouseCursor to a new one. If the new cursor is null,
	it will be reset to the default mouse cursor.
	*/
	this.setMouseCursor = function(newCursor){
		if(newCursor == null){
			fd.canvas.getElement().style.cursor = '';
		}else{
			fd.canvas.getElement().style.cursor = newCursor;
		}
	}
	
	/**
	Converts the Graph to dot language, returning it as a string.
	*/
	this.graphToDot = function(filename){
		var node;
		var dotGraph = 'digraph structs {\nnode [shape=plaintext];\n\n'
		for(var n=0; n < json.length; n++){
			node = json[n];
			if(node.data.$type == "nodeFamily"){
				dotGraph += nodeToDot(node);
				dotGraph += '\n';
			}
		}
		dotGraph += '}';
		console.log(dotGraph);
	}
	
	/**
		Converts a single node to a node in dot language, and returns it as a string.
		@param nodeFamily - the node which is to be converted
	*/
	this.nodeToDot = function(nodeFamily){
		var pWidth = (nodeFamily.data.$width+vertexLabelSize)/2;
		var tableDef = '<TABLE BORDER="0" CELLSPACING="0" CELLBORDER="1" CELLPADDING="0">\n',
			portDefA = '<TR><TD WIDTH="'+pWidth+'" FIXEDSIZE="true" HEIGHT="'+(2*vertexLabelSize)+'" PORT="',
			portDefB = '">p</TD></TR>\n',
			padDefA = '<TR><TD FIXEDSIZE="true" WIDTH="'+pWidth+'" HEIGHT="',
			padDefB = '"></TD></TR>\n',
			emptyDef = '<TD FIXEDSIZE="true" HEIGHT="1" WIDTH="'+pWidth+'"></TD>\n';
		
		// this variable will contain a string, representing a node in dot-code
		var dot = nodeFamily.id,
			dotEdges = '';
		
		dot+= ' [label=< '+tableDef;
		dot+= '<TR><TD COLSPAN="2" HEIGHT="'+(2.5*vertexLabelSize)+'"></TD></TR><TR>\n';
		
		var portIndex = nodeFamily.data.$jsonIndex+2,
			lastPortIndex = nodeFamily.data.$jsonIndex+ nodeFamily.data.$portCount;
		var port = json[portIndex];
		if(port.data.$type == "inputPort"){
			
			dot += '<TD>'+tableDef;
			var dotInput = '',
				countInput = 0;
			
			// write the dot representation of ports into a temporary string
			while(portIndex <= lastPortIndex && port.data.$type == "inputPort"){
				dotInput += portDefA+ port.id + portDefB;
				countInput++;
				portIndex++;
				port = json[portIndex];
			}
			
			// apply padding if there is space above and below the ports
			var padding = nodeFamily.data.$height/2 - (1.5+countInput)*vertexLabelSize;
			if(padding > 0){
				dotInput = padDefA+padding+padDefB + dotInput + padDefA+padding+padDefB;
			}
			
			// append temp string to result
			dot += dotInput + '</TABLE></TD>';
		}
		else{
			dot += emptyDef;
		}
		
		if (portIndex <= lastPortIndex){
			// we get here if there were no input ports OR if we handled them all
			var dotRight = '',
				countRight = 0,
				adja,			// outgoing edges of port
				adjaIter;		// used to iterate through the edges
				
			// add repository ports
			if(port.data.$type == "repositoryPort"){
				
				dot += '<TD>'+tableDef;
				
				// write the dot representation of ports into a temporary string
				while(portIndex <= lastPortIndex && port.data.$type == "repositoryPort"){
					dotRight += portDefA+ port.id + portDefB;
					
					// add edges
					adja = port.adjacencies;
					for(a = 0; a < adja.length; a++){
						adjaIter = adja[a];
						dotEdges += nodeFamily.id+":"+port.id + ":e->" + adjaIter.data.$targetFamily + ":" + adjaIter.nodeTo+":w;\n";
					}
					
					countRight++;
					portIndex++;
					port = json[portIndex];
				}
			}
			// add output ports
			if(portIndex <= lastPortIndex && port.data.$type == "outputPort"){
				
				// if we have preceding repo ports, insert a little gap to separate
				// the output ports
				if(json[portIndex-1].data.$type == "repositoryPort"){
					dotRight += padDefA + vertexLabelSize + padDefB;
					countRight += 0.5;
				}
				else{
					dot += '<TD>'+tableDef;
				}
				
				// write the dot representation of ports into a temporary string
				while(portIndex <= lastPortIndex && port.data.$type == "outputPort"){
					dotRight += portDefA+ port.id + portDefB;
					
					// add edges
					adja = port.adjacencies;
					for(var a = 0; a < adja.length; a++){
						adjaIter = adja[a];
						dotEdges += nodeFamily.id+":"+port.id + ":e->" + adjaIter.data.$targetFamily + ":" + adjaIter.nodeTo+":w;\n";
					}
					
					countRight++;
					portIndex++;
					port = json[portIndex];
				}
			} 
			// apply padding  to right sideif there is space above and below the ports
				var padding = nodeFamily.data.$height/2 - (1.5+countRight)*vertexLabelSize;
				if(padding > 0){
					dotRight = padDefA+padding+padDefB + dotRight + padDefA+padding+padDefB;
				}
				dot += dotRight + '</TABLE></TD>';
		}else{
			dot += emptyDef;
		}
		// end of ports
		
		dot+= '</TR><TR><TD COLSPAN="2" HEIGHT="'+vertexLabelSize/2+'"></TD></TR></TABLE>>];\n';
		dot+= dotEdges;
		
		return dot;
	}
	
	/** 
		reloads the graph, trying to restore all positions 
		this function needs to be called whenever nodes or egdges are
		added or deleted!
	*/
	this.refresh = function(){
		// reload graph
		var mouseIndex = jsonCapacity.max;
		
		fd.loadGraphFlowJSON(json, mouseIndex);
		
		if(grid.visible){
			fd.graph.addNode(grid);
		}
		fd.plot();
		
		// set the mouseNode to the freshly created graph node
		mouseNode = fd.graph.getNode(json[mouseIndex].id);
	}
	
	/**
		Adds a functionality to the following MouseEvents:
			onClick,
			onRightClick,
			onMouseMove,
			onMouseEnter,
			onMouseLeave,
			onDragStart,
			onDragMove,
			onDragCancel,
			onDragEnd
		Or one of these events:
			onCreateEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onCreateNode(node) : boolean,
			onRemoveEdge(sourceNode, targetNode, sourcePort, targetPort) : boolean,
			onRemoveNode(node) : boolean
		When altering the graph, do not forget to refresh() it at the
		end!
		@param event - the name of the mouse event ( see list above)
		@param listenerFunction - a function that may alter the graph
	*/
	this.addListener = function(event, listenerFunction){
		if(listener[event] == undefined){
			listener[event] = [];
		}
		listener[event].push(listenerFunction);
	}
	
	/**
		Calls all listener that are registered under the eventName.
		@param eventName - see 'addListener()'
		@param arguments - an array of arguments. For mouseEvents it is [node, eventInfo, e],
						   for nodeEvents it is [node], 
						   and for edgeEvents it is [sourceNode, targetNode, sourcePort, targetPort]
		@return true if an editor action is permitted
	*/
	function callListener(eventName, arguments){
		var lArr = listener[eventName];
		
		if(!lArr){
			return true;
		}
		
		var permitted = true;
		var testval;
		
		// iterate through all listener this event concerns
		
		// mouse listener functions with 3 arguments (node, eventInfo, e)
		if(arguments.length == 3){
			for(var l=0; l < lArr.length; l++){
				lArr[l](arguments[0], arguments[1], arguments[2]);
			}
			
		// node listener functions with 1 argument (node)
		}else if(arguments.length == 1){
			for(var l=0; l < lArr.length; l++){
				testval = lArr[l](arguments[0]);
				if(testval == false){
					permitted = false;
				}
			}
		
		// edge listener functions with 4 arguments (sourceNode, targetNode, sourcePort, targetPort)
		}else if(arguments.length == 4){
			for(var l=0; l < lArr.length; l++){
				testval = lArr[l](arguments[0], arguments[1], arguments[2], arguments[3]);
				if(testval == false){
					permitted = false;
				}
			}
		}
		return permitted;
	}
	
	/**
	Adds a listener that prohibits edge creation if...
		... an edge with the same source and target exists.
		...	the target node is the source node.
		... the source and target node are of the same port type.
		... a repository node is not connected to a repository and
			vice versa.
	*/
	this.addEdgeConstraints = function(){
		addListener("onCreateEdge", 
		function(sourceFamily, targetFamily, sourcePort, targetPort){
		
			// remove Edge if it does not lead to an inputPort
			if(targetPort.data.$type != "inputPort"){
				return false;
			}
		
			// remove Edge if the connected nodes share the same type
			if(sourcePort.data.$type == targetPort.data.$type){
				return false;
			}
			
			var goesToRepo = (targetFamily.data.$nodeType == "Repository"),
				isRepoPort = (sourcePort.data.$type == "repositoryPort");
			
			// remove Edge if it is falsely connected to repository ports
			if(goesToRepo != isRepoPort){
				return false;
			}
			
			// remove duplicate Edge
			var adja = sourcePort.adjacencies,
				duplicate = false;
			for(var a=0; a < adja.length; a++){
				if(adja[a].nodeTo == targetPort.id){
					if(duplicate){
						return false;
					}
					else{
						duplicate = true;
					}
				}
			}
			return true;
		});
	}
	
	/**
	Iterates through all nodes and ports, calling a given function for each.
	@param nodeFunction - a single parameter function, called on a node.
	*/
	this.iterateAllNodes = function(nodeFunction){
		var node;
		for(var n=0, l=jsonCapacity.occupied; n < l; n++){
			node = json[n];
			nodeFunction(node);
		}
		for(var n=jsonCapacity.max+1, l= json.length; n < l; n++){
			node = json[n];
			nodeFunction(node);
		}
	}
	
	/**
	Changes the colors of the entire graph.
	*/
	this.setNodeStyle = function(fillColor, strokeColor, portColor){
		nodeFillColor['nodeFamily'] = fillColor;
		nodeStrokeColor['nodeFamily'] = strokeColor;
		nodeStrokeColor['inputPort'] = portColor;
		nodeStrokeColor['outputPort'] = portColor;
		nodeStrokeColor['repositoryPort'] = portColor;
		
		var data;
		// change family colors
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
			data = json[n].data;
			data.$fillColor = nodeFillColor["nodeFamily"];
			data.$color = nodeStrokeColor["nodeFamily"];
		}
		// change port colors
		for(var n = jsonCapacity.max+1, l = json.length; n < l; n++){
			data = json[n].data;
			data.$color = nodeStrokeColor[data.$type];
			if(data.$type == "crossBox"){
				data.$fillColor = nodeFillColor["nodeFamily"];
			}
		}
		refresh();
	}
	
	/**
	Permits or prohibits node and edge deletion/creation and hides the cross
	accordingly.
	@param deleteCreate - if true, one can change the graph in any way
	@param move - if true, one can change the graph in any way
	*/
	this.setReadOnly = function(deleteCreate, move){
		
		if(deleteCreate != readOnly.deleteCreate){
			// set crossBoxes (in)visible
			var data;
			for(var n=jsonCapacity.max+1, l= json.length; n < l; n++){
				data = json[n].data;
				if(data.$type == "crossBox"){
					data.$visible = deleteCreate;
				}
			}
			readOnly.deleteCreate = deleteCreate;
		}
		
		if(move != undefined){
			readOnly.move = move;
		}
	}
		
	/**
	 Adds a Node to the graph.
	 @param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	 @param inputPortIDs - an array of properties for the input ports (id, name, tooltip)
	 @param outputPortIDs - an array of properties for the output ports (id, name, tooltip)
	 */
	this.addNode = function(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts, nodeType, forced){
	
	// check if we have enough space reserved
	var max = jsonCapacity.max;
	var occ = jsonCapacity.occupied;
	if(occ == max){
		var families = json.slice(0, max);
		var ports = json.slice(max);
		
		for(var n = 0; n < max; n++){
			families.push(null);
		}
		max *= 2;
		jsonCapacity.max = max;
		json = families.concat(ports);
	}
	
	var countRepo = 0,
		countInput = 0,
		countOutput = 0,
		iconSpace = 0;
		
	if(repositoryPorts != null){
		countRepo= repositoryPorts.length;
	}
	if(inputPorts != null){
		countInput= inputPorts.length;
	}
	if(outputPorts != null){
		countOutput= outputPorts.length;
	}
	
	var maxPorts = Math.max(countInput, countRepo+countOutput+0.5),
		size = vertexLabelSize*2;
	
	// insert some space between repo ports and output ports, only if both are present
	if(maxPorts > countInput && (countRepo == 0 || countOutput == 0)){
		maxPorts-= 0.5;
	}
	
	// check if this nodeType has a defined icon
	if(nodeIcons[nodeType] != undefined){
		iconSpace = 6;
	}
		
	var width = (nodeFamily.name.length+4+iconSpace)*vertexLabelSize,
		height = size + 
			Math.max( 2*size, size/2 + maxPorts * size);
	
	// the classname may be longer than the diplayed name
	if( nodeFamily.nodeClass != undefined){
		width = Math.max(width,  (nodeFamily.nodeClass.length+6+iconSpace)*(vertexLabelSize-2));
	}
	
	// add big node box
	// change node color, depending on its type
	var nodeColor = nodeFillColor["nodeFamily"],
		strokeColor = nodeStrokeColor["nodeFamily"];
	var newNode = {
		  "data": {
			"$dim": size,					
			"$type": "nodeFamily",
			"$jsonIndex": occ,
			"$portCount": countRepo+countInput+countOutput,
			"$portIndex": json.length - max,
			"$icon": nodeIcons[nodeType],
			"$width": width,
			"$height": height,
			"$color": strokeColor,
			"$fillColor": nodeColor,
			"$nodeClass": nodeFamily.nodeClass,
			"$nodeType": nodeType,
			"$xPos": xPosition,
			"$yPos": yPosition,
			"$tooltip": nodeFamily.tooltip
		  }, 
		  "id": nodeFamily.id, 
		  "name": nodeFamily.name
	};
	json[occ] = newNode;
	
	// add closeButton
		var closeButton ={
			  "adjacencies": [], 
			  "data": {
				"$dim": size,					
				"$type": "crossBox",
				"$family": newNode,
				"$xPos": xPosition+ width/2,
				"$yPos": yPosition- height/2,
				"$relativeX": width/2,
				"$relativeY": -height/2,
				"$color": strokeColor,
				"$fillColor": nodeColor,
				"$visible":true
			   }, 
			  "id": nodeFamily.id+".close", 
			  "name": "x"
		};
	json.push(closeButton);
	json[occ].data.$portCount++;
	
	// Loop Info: 
	// This loop adds all port types to the temporary port array.
	// For fast position computation, the relative position
	// of the first port to the nodeFamily-box is computed for each port type
	// it is stored in relativeX, relativeY
	
	for(var portType = 0; portType < 3; portType++){
		var x, y, relativeX, relativeY,
			loopPorts, loopType;
		switch(portType){
			case 0: // input ports
				relativeX = -width/2;
				relativeY = (0.75 - countInput/2 )*size;
				loopPorts = inputPorts;
				loopType = "inputPort";
				break;
		
			case 1: // repository ports
				var repoSpace = 0.5;
				if(countOutput == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput+countRepo+repoSpace)/2 )*size;
				loopPorts = repositoryPorts;
				loopType = "repositoryPort";
				break;
				
			case 2: // output ports
				var repoSpace = 0.5;
				if(countRepo == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput-countRepo-repoSpace)/2 )*size;
				loopPorts = outputPorts;
				loopType = "outputPort";
				break;
		}
		
		// continue only if ports exist
		if(loopPorts != null && loopPorts.length > 0){
		
			x = xPosition+ relativeX;
			y = yPosition+ relativeY;
			
			for(var p=0; p < loopPorts.length; p++){
				var port = loopPorts[p];
				var newPort ={
					  "adjacencies": [], 
					  "data": {
						"$dim": size,					
						"$type": loopType,
						"$xPos": x,
						"$yPos": y,
						"$color": nodeStrokeColor[loopType],
						"$fillColor": nodeColor,
						"$tooltip": port.tooltip}, 
					  "id": nodeFamily.id+"."+port.id, 
					  "name": port.name};
					  
				if(p == 0){
					newPort.data.$relativeX= relativeX;
					newPort.data.$relativeY= relativeY;
				}
				json.push(newPort);
				y += size;
			}
		}
	}
	
		// call listener
		if(!forced){
			callListener("onCreateNode", [newNode]);
		}
		jsonCapacity.occupied++;
	}

	/**
		Removes a Node and all its edges from the graph.
		@param nodeFamily - the node which is to be removed
	*/
	this.removeNode =  function(nodeFamily, forced){
		//fd.graph.removeNode(node.id);
		// call listener
		if(!forced && !callListener("onRemoveNode", [nodeFamily])){
			return;
		}
		
		var deleteFrom = nodeFamily.data.$portIndex + jsonCapacity.max, 
			deleteUntil = deleteFrom+ nodeFamily.data.$portCount,
			familyIndex = nodeFamily.data.$jsonIndex,
			occ = --jsonCapacity.occupied; // decrement occupied spaces
		
		// delete nodeFamily and close button
		delete json[familyIndex];
		delete json[deleteFrom];
		
		// delete ports
		for(var n = deleteFrom+1;  n < deleteUntil;  n++){
			var portID = json[n].id;
			
			// remove all incoming edges
			for(var e = 0;  e < json.length;  e++){
				var incNode = json[e];
				if(incNode != undefined){
					removeEdge(incNode.id, portID, true);	
				}
			}
			
			delete json[n];
		}
		
		// remove space that was used for ports
		var deletionSum = deleteUntil-deleteFrom; 
		json.splice(deleteFrom, deletionSum);
		selectedNode = null;
		
		/*
		// update nodeFamily indices
		for(var n = deleteUntil-deletionSum+1;  n < json.length;  n++){
			var familyIndex = json[n].data.$jsonIndex;
			if(familyIndex != undefined){
				json[n].data.$jsonIndex = n;
			}
		}
		*/
		
		// update nodeFamily indices and remove the deletion gap
		// in the array space by shifting all following families up
		var data;
		for(var n = familyIndex;  n < occ;  n++){
			json[n] = json[n+1];
			data = json[n].data;
			data.$jsonIndex--
			data.$portIndex-= deletionSum;
			
		}
		json[occ] = null;
		
		
		// check if we should remove some space
		var max = jsonCapacity.max;
		if( (occ*4) < max && max > 10){
			var ports = json.slice(max);
			
			max /= 2;
			jsonCapacity.max = max;
			var families = json.slice(0, max);
			json = families.concat(ports);
		}
		return;
	 }
	 
	 /**
		Prepares a drag opertion for a node, by memorizing all corresponding graph nodes.
		This speeds up the moveNode() function considerably.
	*/
	 function prepareNodeMove(node){
		var nodes = [];
		var familyIndex = node.data.$jsonIndex,
			moveFrom = node.data.$portIndex+ jsonCapacity.max, 
			moveUntil = moveFrom+ node.data.$portCount;
		
		// memorize the nodeFamily
		nodes.push(node);
		
		// memorize all ports and the crossbox
		var subnode;
		for(var i= moveFrom; i< moveUntil; i++){
			subnode = fd.graph.getNode(json[i].id);
			nodes.push(subnode);
			
		}
		
		navi.draggedNodes = nodes;
	 }
	 
	 /**
		Moves a nodeFamily and its subnodes to the specified position.
	 */
	 function moveNode(xPos, yPos){
		var nodes = navi.draggedNodes,
			node = nodes[0],
			dim = node.data.$dim;
		var x = xPos,
			y = yPos;
			
		// snap to grid
		if(grid.snap){
			var gridDim = grid.data.$dim,
				left = (xPos - node.data.$width/2),
				top = (yPos - node.data.$height/2),
				right = (xPos + node.data.$width/2),
				bottom = (yPos + node.data.$height/2),
				
				offLeft = Math.abs(left % gridDim),
				offTop = Math.abs(top % gridDim),
				offRight = Math.abs(right % gridDim),
				offBottom = Math.abs(bottom % gridDim),
				threshold = gridDim/6;
		
			if(top > 0){
				offTop = gridDim - offTop;
			}
			if(bottom > 0){
				offBottom = gridDim - offBottom;
			}
			if(left > 0){
				offLeft = gridDim - offLeft;
			}
			if(right > 0){
				offRight = gridDim - offRight;
			}
			
			// snap to closest gridline
			if(offLeft < threshold) {
				x += offLeft;
			}
			if(offRight < threshold) {
				x += offRight;
			}
			else if(offLeft > gridDim - threshold){
				x -= gridDim-offLeft;
			}
			else if(offRight > gridDim - threshold){
				x -= gridDim-offRight;
			}
			
			if(offTop < threshold) {
				y += offTop;
			}
			
			else if(offTop > gridDim - threshold){
				y -= gridDim-offTop;
			}
			
			else if(offBottom < threshold) {
				y += offBottom;
			}
			
			else if(offBottom > gridDim - threshold){
				y -= gridDim-offBottom;
			}
		}
	 
	 
		// set position of nodeFamily
		node.pos.setc(x, y);
		
		// set position of ports and crossbox
		var nodeX, nodeY;
		for(var i= 1, l = nodes.length; i < l; i++){
			node = nodes[i];
			
			if(node.data.$relativeX){
				nodeX = x+ node.data.$relativeX;
				nodeY = y+ node.data.$relativeY;
			}
			node.pos.setc(nodeX,nodeY);
			nodeY += dim;
		}
	 }
	 
	 /**
		Saves the position of a node, so that it will be placed there again,
		when the graph is altered.
	 */
	 function saveNodeMove(){
		var node = navi.draggedNodes[0],
			familyIndex = node.data.$jsonIndex,
			moveFrom = node.data.$portIndex+ jsonCapacity.max-1, 
			nodes = navi.draggedNodes,
			pos = node.pos.getc(true);
		
		// set pos of family box
		var data= json[familyIndex].data;
		data.$xPos = pos.x;
		data.$yPos = pos.y;
		
		// set pos of ports
		for(var i= 1, l = nodes.length; i < l; i++){
			data = json[moveFrom+i].data;
			pos = nodes[i].pos.getc(true);
			
			data.$xPos = pos.x;
			data.$yPos = pos.y;
		}
		navi.draggedNodes = null;
	 }

  
  /**
   *  adds an edge to the graph
   *  @param sourceID the id of the node from which the edge starts
   *  @param targetID the id of the node where the edge ends
   *  @return true if the edge was successfully added
   */
	this.addEdge = function(sourceID, targetID, edgeLabel, forced){
		
		// look up the source node and the nodeFamilys of both nodes
		var source, sourceFamily, targetFamily, target;
		
		var sourceFamilyID = sourceID.substring(0, sourceID.indexOf(".")), 
			targetFamilyID = targetID.substring(0, targetID.indexOf(".")),
			loopNode;
		// look for nodefamilies
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){
	
			loopNode = json[n];
			
			if(loopNode.id == sourceFamilyID){
				sourceFamily = loopNode
			}
			if(loopNode.id == targetFamilyID){
				targetFamily = loopNode
			}
		} 
		
		// look for ports
		for(var n = jsonCapacity.max, l = json.length; n < l; n++){
			loopNode = json[n];
			
			if(loopNode.id == sourceID){
				source = loopNode;
			}
			if(loopNode.id == targetID){
				target = loopNode;
			}
		}
		// call listener if the edge is not connected to the dummy node and check for permission
		if(!forced && sourceID != mouseNode.id && targetID != mouseNode.id){
			if (!callListener("onCreateEdge", [sourceFamily, targetFamily, source, target])){
				return false;
			}
		}
		
		// if edge is connected to mouse, there is no targetFamily...
		var targetFamilyID = null;
		if(targetFamily != undefined){
			targetFamilyID = targetFamily.id;
		}
		
		var adja = source.adjacencies;
		// source node does not exist?
		if(adja == undefined){
			return false;
		}
	
		var edge ={
				  "nodeTo": targetID, 
				  "nodeFrom": sourceID,
				  "data": {"$direction" : [ sourceID, targetID ],
						   "$targetFamily" : targetFamilyID},
				  };
		
		if(edgeLabel != null && edgeLabel != undefined){
			edge.data.$label = edgeLabel;
        }
				  
		adja.push(edge);
		
		if(sourceID == mouseNode.id || targetID == mouseNode.id){
			edge.data.$type = "mouseArrow";
		}
		
		return true;
	}
  
  /**
	Removes the edge between two vertices.
	@param source the source Node
	@param targetID the id of the target Node
	*/
  this.removeEdge = function(sourceID, targetID, forced){
	// look up the source node and the nodeFamilys of both nodes
	var source, sourceFamily, targetFamily, target;
	
	var sourceFamilyID = sourceID.substring(0, sourceID.indexOf(".")), 
		targetFamilyID = targetID.substring(0, targetID.indexOf(".")),
		loopNode;
		
	// look for ports
	for(var n = jsonCapacity.max, l = json.length; n < l; n++){
		loopNode = json[n];
		if(loopNode != undefined){
			if(loopNode.id == sourceID){
				source = loopNode;
			}
			if(loopNode.id == targetID){
				target = loopNode;
			}
		}
	}
		
	if(!forced){
		// look for nodefamilies
		for(var n = 0, l = jsonCapacity.occupied; n < l; n++){

			loopNode = json[n];
			
			if(loopNode.id == sourceFamilyID){
				sourceFamily = loopNode
			}
			if(loopNode.id == targetFamilyID){
				targetFamily = loopNode
			}
		} 
	
		// call listener and ask for permission to delete
		if(sourceID != mouseNode.id && targetID != mouseNode.id){
			if(!callListener("onRemoveEdge", [sourceFamily, targetFamily, source, target])){
				return;
			}
		}
	}
	
	// nothing to delete?
	if(source == undefined){
		return;
	}
	var adja = source.adjacencies;
	
	for(var a=0; a< adja.length; a++){
		if(adja[a].nodeTo == targetID){
			delete adja[a];
			adja.splice(a,1);
			break;
		}
	}
	
  }
  
  /**
	makes it so that a text label is not selectable.
	(unfortunately, this does not prevent CTRL+A selection)
	*/
  function setLabelUnselectable(label){
	// internet explorer:
	if (typeof label.onselectstart!="undefined"){
        label.onselectstart =
			function(){return false};
		}
	// mozilla firefox:
    else if (typeof label.style.MozUserSelect!="undefined"){
        label.style.MozUserSelect = "none";
	}
	// others:
    else{
        label.onmousedown =
			function(){return false};
	}
  }
  
  /**
  returns a jsonNode by id
  @param nodeID the id of the desired node
  */
  this.getNode = function(nodeID){
	for(var n=0; n< json.length; n++){
		var node = json[n];
		if(node != undefined && node.id == nodeID){
			return json[n];
		}
	}
	return null;
  }
  
  /**
	clears all highlights, caused by mouse-enter events
	@param clearNodes - if true, highlighted nodes will be cleared
	@param clearEdges - if true, highlighted edges will be cleared
	*/
  function setHighlight(node){
		// animation parameters
		var trans = $jit.Trans.Quint.easeOut, 
			dur= 200;
		// clean up the old highlight
		if(hover != null){
			
			if(hover.nodeFrom != undefined){
				hover.setData('lineWidth', 1, 'end');
				hover.setData('color', edgeColor, 'end');
			}
			else{
				var type = hover.data.$type;
				hover.setData('color', nodeStrokeColor[type], 'end');
			}
			hover = null;
		}
		
		// do we highlight something new?
		if(node != null){
			if(node.nodeFrom != undefined){
				node.setData('color', edgeColorFocus, 'end');
				node.setData('lineWidth', 2, 'end');
			}else{
				node.setData('color', nodeColorFocus, 'end');
			}
			hover = node;
		}else{
			// slower fade away transition
			trans = $jit.Trans.Quint.easeIn;
			dur = 400;
		}
		
		// Animations overwrite each other. If we let this animation
		// execute while we delete an edge, the onComplete() function
		// of the edge animation will not be called
		fd.animate({  
				modes: ['edge-property:lineWidth:color',
						'node-property:color'],
				transition: trans,  
				duration: dur
		}); 
  }
    
  /**
	Initializes the Graph
	*/
  this.initGraph = function(){
	json = [ null, null, null, null, null, null, null, null, null, null, 
			{
			  "adjacencies": [], 
			  "data": {
				"$dim": 0,
				"$type": "none",
				"&xPos": 0,
				"&typeSelected" : null,
				"&yPos": 0,
			  }, 
			  "id": "#DUMMY_MOUSE_NODE", 
			  "name": "",
			  "alpha": 0
		}];
	  
	  fd = new $jit.FlowGraph({
		//id of the visualization container
		injectInto: visContainer,
		//Enable zooming and panning
		//with scrolling and DnD
		Navigation: {
		  enable: true,
		  //Enable panning events only if we're dragging the empty
		  //canvas (and not a node).
		  panning: 'avoid nodes',
		  zooming: 40 //zoom speed. higher is more sensible
		},
		// Change node and edge styles such as
		// color and width.
		// These properties are also set per node
		// with dollar prefixed data-properties in the
		// JSON structure.
		Node: {
		  overridable: true,
		  CanvasStyles: {  
			  lineWidth: 2  
			}  
		  
		},
		Edge: {
		  overridable: true,
		  color: edgeColor,
		  lineWidth: 1,
		  dim: vertexLabelSize,
		  type: 'flowArrow'
		},
		//Native canvas text styling  
		Label: {  
			type: 'HTML', //Native or HTML  
			size: 10,  
			style: 'bold'  
		}, 
		//Add Tips  
		  Tips: {  
			enable: true,  
			type: 'Native',
			onShow: function(tip, node) {  
			  var message = node.data.$tooltip;
			  
			  // do not show a tooltip if it is undefined
			  if(message == undefined){
				tip.style.visibility = "hidden";
			  }
			  else{
				tip.innerHTML = "<div class=\"tip-text\">" + message + "</div>";  
				tip.style.visibility = "visible";
			  }
			}  
		 },  
		// Add node events
		Events: {
		  enable: true,
		  enableForEdges: true,
		  type: 'Native',
		  
		  /**
		   *	EVENT:	OnRightClick
		   */
		  onRightClick: function(node, eventInfo, e) {
			callListener("onRightClick", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragCancel
		   */
		  onDragCancel: function(node, eventInfo, e) {
			callListener("onDragCancel", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseEnter
		   */
		  onMouseEnter: function(node, eventInfo, e) {
			if(node.id == mouseNode.id){
				return;
			}
			
			// ignore some highlights if we are dragging an edge
			if(selectedNode != null && 
				(node.data.$type != "inputPort" 
				&& node.data.$type != "outputPort" 
				&& node.data.$type != "repositoryPort")){
				return;
			}
			
			// check if object is an edge or a node
			mouseOverNode = (node.nodeFrom == undefined);
			mouseOverEdge = !mouseOverNode;
			
			setHighlight(node);
			if(readOnly.move && node.data.$type == "nodeFamily"){
				fd.canvas.getElement().style.cursor = 'move';
			}else if (readOnly.deleteCreate){
				fd.canvas.getElement().style.cursor = 'pointer';
			}
			callListener("onMouseEnter", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseMove
		   */
		  onMouseMove: function(node, eventInfo, e) {
		  /*if(!node && navi.isDragging){
			var mousePos = eventInfo.getPos(),
				deltaX = mousePos.x - navi.dragX,
				deltaY = mousePos.y - navi.dragY;
			fd.canvas.translate(deltaX, deltaY);
			deltaX = mousePos.x;
			deltaY = mousePos.y;
		  }
		  */
			// are we dragging an edge with the mouse ?
			if(selectedNode != null){
				var pos = eventInfo.getPos();
				mouseNode.pos.setc(pos.x, pos.y-2);
				fd.plot();
			}
			callListener("onMouseMove", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseLeave
		   */
		  onMouseLeave: function(node, eventInfo, e) {
			fd.canvas.getElement().style.cursor = '';
			mouseOverNode = false;
			mouseOverEdge = false;
			
			setHighlight(null);
			callListener("onMouseLeave", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragStart
		   */
		  onDragStart: function(node, eventInfo, e) {
		  var mousePos = eventInfo.getPos();
		  
			if(node == false || node.id == mouseNode.id ||
			   node.data.$type != "nodeFamily"){
			    navi.dragX = e.pageX;
				navi.dragY = e.pageY;
				navi.isDragging = true;
				return;
			}
			
			// memorize where we drag the node instead on centering
			// it to the mouse pointer
			var	nodePos = node.pos.getc(true);
				
			navi.dragX = nodePos.x - mousePos.x;
			navi.dragY = nodePos.y - mousePos.y;
			
			// we memorize all subnodes of the dragged nodefamily, to speed up
			// dragging
			prepareNodeMove(node);
			
			callListener("onDragStart", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragMove
		   */
		  onDragMove: function(node, eventInfo, e) {
			if(!readOnly.move || !node || 
				node.id == mouseNode.id || node.data.$type != "nodeFamily"){
					return;
			}
			var pos = eventInfo.getPos();
			moveNode(pos.x+navi.dragX,
					 pos.y+navi.dragY);
			callListener("onDragMove", [node, eventInfo, e]);
			fd.plot();
			
		  },
		  
		  /**
		   *	EVENT:	OnDragEnd
		   */
		  onDragEnd: function(node, eventInfo, e) {
			if(node.data.$type == "nodeFamily"){
				saveNodeMove();
			}
			fd.plot();
			callListener("onDragEnd", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnTouchMove
		   */
		  onTouchMove: function(node, eventInfo, e) {
			$jit.util.event.stop(e); //stop default touchmove event
			this.onDragMove(node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnClick
		   */
		  onClick: function(node, eventInfo, e){
		  
		  if(node == false){
			var deltaX = e.pageX - navi.dragX,
				deltaY = e.pageY - navi.dragY;
			navi.centerX -= deltaX;
			navi.centerY -= deltaY;
			//fd.canvas.translate(deltaX, deltaY);
		  }
		  else if(mouseOverNode && readOnly.deleteCreate){
		  
			switch(node.data.$type){
				case "crossBox":
					removeNode(node.data.$family);
					refresh();
					break;
					
				case "nodeFamily":
					break;
					
				case "outputPort":
				case "inputPort":
				case "repositoryPort":
					// no selection yet
					if(selectedNode == null){
						selectedNode = {"id": node.id, "from": node.data.$type};
						
						// add edge to mouseNode
						if(selectedNode.from == "inputPort"){
							addEdge(mouseNode.id, node.id);
						}
						else{
							addEdge(node.id, mouseNode.id);
						}
						refresh();
						return;
					}
					// click outside the edge drag box
					else{
						// add Edge if the selectedNode differs from the clickedNode
						var newEdgeAdded;
						if(selectedNode.from == "inputPort"){
							var label = selectedNode.label;
							newEdgeAdded = addEdge(node.id, selectedNode.id, label);
						}
						else{
							var label = selectedNode.label;
							newEdgeAdded = addEdge(selectedNode.id, node.id, label);
						}
						if(newEdgeAdded){
							refresh();
						}
					}
					break;
			}
			
		  }
		  else if(mouseOverEdge && readOnly.deleteCreate){
			// no selection yet
				if(selectedNode == null){
					selectedNode = {"id":node.data.$direction[0], "from":"outputPort", "label":node.data.$label};//nodeTo
					// remove selectedEdge
					var label = node.data.$label;
					removeEdge(selectedNode.id, node.data.$direction[1]);
					// add edge to mouseNode
					addEdge(selectedNode.id, mouseNode.id, label);
					refresh();
					return;
				}
		  }
		  
		  // remove selection and mouse-edge upon clicking anywhere
		  if(selectedNode != null){
		  
			if(selectedNode.from == "inputPort"){
				removeEdge(mouseNode.id, selectedNode.id, true);
			}else{
				removeEdge(selectedNode.id, mouseNode.id, true);
			}
			
			selectedNode = null;
			refresh();
		  }
		  callListener("onClick", [node, eventInfo, e]);
		  navi.isDragging = false;
		  }
		},
		
		
		//Number of iterations for the FD algorithm
		iterations: 0,
		//Edge length
		levelDistance: 0,
		// This method is only triggered
		// on label creation and only for DOM labels (not native canvas ones).
		/*onCreateLabel: function(domElement, node){
			// do not add labels if it is a dummy element
			// instead save pointer to domElement
			if(node.id == "#DUMMY_MOUSE_NODE"){
				mouseNode = node;
				return;
			}
			if(node.data.$type != "nodeFamily"){
				return;
			}
			
			domElement.innerHTML = node.name;  
			
			// set style for name
			var style = domElement.style;  
			style.fontSize = 1.65* fd.canvas.scaleOffsetX* vertexLabelSize+"px";
			style.left = 0.2*vertexLabelSize-domElement.offsetWidth/2+"px";
			setLabelUnselectable(domElement);
		},
		
		// Change node styles when DOM labels are placed
		// or moved.
		onPlaceLabel: function(domElement, node){
		  if(node.data.$type != "nodeFamily"){
			return;
		  }
		  
		  var style = domElement.style;
		  var left = parseInt(style.left);
		  var top = parseInt(style.top);
		  var scale = fd.canvas.scaleOffsetX;
		  var w = scale*vertexLabelSize* node.name.length;
		  
		  var classWidth =  scale*(vertexLabelSize-2)* (node.data.$nodeClass.length+2);
		  if(classWidth > w){
			w = classWidth;
		  }
		  
		  style.fontSize = scale* 1.65 * vertexLabelSize+"px";
		  style.fontFamily = "Lucida Console";
		  style.left = (left - w/ 2) + 'px';
		  style.top = (top - h ) + 'px';
		  style.color = node.data.$color;
		  style.display = '';
		  // add class name
			if(node.data.$nodeClass != undefined){
				domElement.innerHTML = node.name;  
				var fontSize= scale*1.65*(vertexLabelSize-2);
				domElement.innerHTML += "<div style=\"font-size:"+fontSize+"px\"> &lt"+node.data.$nodeClass+"&gt";
			}
		}*/
	  });
	  
	  
	  // load JSON data.
	  fd.loadGraphFlowJSON(json,jsonCapacity.max);
	  // compute positions
	  refresh();
	  // end
	}
	
	// initialize graph
	initGraph();
	return this;
}